
/*  A Bison parser, made from ftpcmd.y
 by  GNU Bison version 1.25
  */

#define YYBISON 1  /* Identify Bison output.  */

#define	A	258
#define	B	259
#define	C	260
#define	E	261
#define	F	262
#define	I	263
#define	L	264
#define	N	265
#define	P	266
#define	R	267
#define	S	268
#define	T	269
#define	SP	270
#define	CRLF	271
#define	COMMA	272
#define	USER	273
#define	PASS	274
#define	ACCT	275
#define	REIN	276
#define	QUIT	277
#define	PORT	278
#define	PASV	279
#define	TYPE	280
#define	STRU	281
#define	MODE	282
#define	RETR	283
#define	STOR	284
#define	APPE	285
#define	MLFL	286
#define	MAIL	287
#define	MSND	288
#define	MSOM	289
#define	MSAM	290
#define	MRSQ	291
#define	MRCP	292
#define	ALLO	293
#define	REST	294
#define	RNFR	295
#define	RNTO	296
#define	ABOR	297
#define	DELE	298
#define	CWD	299
#define	LIST	300
#define	NLST	301
#define	SITE	302
#define	STAT	303
#define	HELP	304
#define	NOOP	305
#define	MKD	306
#define	RMD	307
#define	PWD	308
#define	CDUP	309
#define	STOU	310
#define	SMNT	311
#define	SYST	312
#define	SIZE	313
#define	MDTM	314
#define	LPSV	315
#define	LPRT	316
#define	SPASV	317
#define	SPORT	318
#define	UMASK	319
#define	IDLE	320
#define	CHMOD	321
#define	LEXERR	322
#define	STRING	323
#define	NUMBER	324

#line 41 "ftpcmd.y"


#ifndef lint
static char sccsid[] = "@(#)ftpcmd.y	8.3 (Berkeley) 4/6/94";
#endif /* not lint */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <netinet/in.h>
#ifdef INET6
#include <netinet6/in6.h>
#endif /* INET6 */
#include <arpa/ftp.h>

#include <ctype.h>
#include <errno.h>
#include <glob.h>
#include <pwd.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include "extern.h"

#ifdef INET6
#include "support.h"

#define su union { \
  struct sockaddr sa; \
  struct sockaddr_in sin; \
  struct sockaddr_in6 sin6; \
}

extern su his_addr;
extern su ctrl_addr;
extern su data_dest;
#else /* INET6 */
extern	struct sockaddr_in data_dest;
#endif /* INET6 */
extern	int logged_in;
extern	struct passwd *pw;
extern	int guest;
extern	int logging;
extern	int type;
extern	int form;
extern	int debug;
extern	int timeout;
extern	int maxtimeout;
extern  int pdata;
extern	char hostname[], remotehost[];
extern	char proctitle[];
extern	int usedefault;
extern  int transflag;
extern  char tmpline[];

off_t	restart_point;

static	int cmd_type;
static	int cmd_form;
static	int cmd_bytesz;
char	cbuf[512];
char	*fromname;

struct tab {
	char	*name;
	short	token;
	short	state;
	short	implemented;	/* 1 if command is implemented */
	char	*help;
};

struct tab cmdtab[], sitetab[];
static void	 help __P((struct tab *, char *));
static void	 sizecmd __P((char *));


#line 125 "ftpcmd.y"
typedef union {
	int	i;
	char   *s;
} YYSTYPE;
#include <stdio.h>

#ifndef __cplusplus
#ifndef __STDC__
#define const
#endif
#endif



#define	YYFINAL		220
#define	YYFLAG		-32768
#define	YYNTBASE	70

#define YYTRANSLATE(x) ((unsigned)(x) <= 324 ? yytranslate[x] : 85)

static const char yytranslate[] = {     0,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     2,     2,     2,     2,     2,     1,     2,     3,     4,     5,
     6,     7,     8,     9,    10,    11,    12,    13,    14,    15,
    16,    17,    18,    19,    20,    21,    22,    23,    24,    25,
    26,    27,    28,    29,    30,    31,    32,    33,    34,    35,
    36,    37,    38,    39,    40,    41,    42,    43,    44,    45,
    46,    47,    48,    49,    50,    51,    52,    53,    54,    55,
    56,    57,    58,    59,    60,    61,    62,    63,    64,    65,
    66,    67,    68,    69
};

#if YYDEBUG != 0
static const short yyprhs[] = {     0,
     0,     1,     4,     7,    12,    17,    23,    29,    35,    38,
    41,    44,    49,    54,    59,    64,    73,    79,    85,    91,
    95,   101,   105,   111,   117,   120,   126,   131,   134,   138,
   144,   147,   152,   155,   161,   167,   171,   175,   180,   187,
   193,   201,   211,   216,   223,   229,   232,   238,   244,   247,
   250,   256,   261,   263,   264,   266,   268,   280,   282,   284,
   286,   288,   292,   294,   298,   300,   302,   306,   309,   311,
   313,   315,   317,   319,   321,   323,   325,   327
};

static const short yyrhs[] = {    -1,
    70,    71,     0,    70,    72,     0,    18,    15,    73,    16,
     0,    19,    15,    74,    16,     0,    23,    84,    15,    76,
    16,     0,    63,    84,    15,    81,    16,     0,    61,    84,
    15,    81,    16,     0,    24,    16,     0,    62,    16,     0,
    60,    16,     0,    25,    15,    78,    16,     0,    26,    15,
    79,    16,     0,    27,    15,    80,    16,     0,    38,    15,
    69,    16,     0,    38,    15,    69,    15,    12,    15,    69,
    16,     0,    28,    84,    15,    81,    16,     0,    29,    84,
    15,    81,    16,     0,    30,    84,    15,    81,    16,     0,
    46,    84,    16,     0,    46,    84,    15,    68,    16,     0,
    45,    84,    16,     0,    45,    84,    15,    81,    16,     0,
    48,    84,    15,    81,    16,     0,    48,    16,     0,    43,
    84,    15,    81,    16,     0,    41,    15,    81,    16,     0,
    42,    16,     0,    44,    84,    16,     0,    44,    84,    15,
    81,    16,     0,    49,    16,     0,    49,    15,    68,    16,
     0,    50,    16,     0,    51,    84,    15,    81,    16,     0,
    52,    84,    15,    81,    16,     0,    53,    84,    16,     0,
    54,    84,    16,     0,    47,    15,    49,    16,     0,    47,
    15,    49,    15,    68,    16,     0,    47,    15,    64,    84,
    16,     0,    47,    15,    64,    84,    15,    83,    16,     0,
    47,    15,    66,    84,    15,    83,    15,    81,    16,     0,
    47,    15,    65,    16,     0,    47,    15,    65,    15,    69,
    16,     0,    55,    84,    15,    81,    16,     0,    57,    16,
     0,    58,    84,    15,    81,    16,     0,    59,    84,    15,
    81,    16,     0,    22,    16,     0,     1,    16,     0,    40,
    84,    15,    81,    16,     0,    39,    15,    75,    16,     0,
    68,     0,     0,    68,     0,    69,     0,    69,    17,    69,
    17,    69,    17,    69,    17,    69,    17,    69,     0,    10,
     0,    14,     0,     5,     0,     3,     0,     3,    15,    77,
     0,     6,     0,     6,    15,    77,     0,     8,     0,     9,
     0,     9,    15,    75,     0,     9,    75,     0,     7,     0,
    12,     0,    11,     0,    13,     0,     4,     0,     5,     0,
    82,     0,    68,     0,    69,     0,     0
};

#endif

#if YYDEBUG != 0
static const short yyrline[] = { 0,
   163,   164,   169,   173,   178,   183,   195,   221,   250,   258,
   264,   270,   308,   320,   332,   336,   340,   347,   354,   361,
   366,   373,   378,   385,   392,   396,   403,   414,   418,   423,
   430,   434,   449,   453,   460,   467,   472,   477,   481,   485,
   495,   510,   524,   530,   544,   551,   572,   589,   610,   615,
   621,   633,   643,   647,   651,   655,   659,   717,   721,   725,
   732,   737,   742,   747,   752,   756,   761,   767,   775,   779,
   783,   790,   794,   798,   805,   837,   841,   868
};
#endif


#if YYDEBUG != 0 || defined (YYERROR_VERBOSE)

static const char * const yytname[] = {   "$","error","$undefined.","A","B",
"C","E","F","I","L","N","P","R","S","T","SP","CRLF","COMMA","USER","PASS","ACCT",
"REIN","QUIT","PORT","PASV","TYPE","STRU","MODE","RETR","STOR","APPE","MLFL",
"MAIL","MSND","MSOM","MSAM","MRSQ","MRCP","ALLO","REST","RNFR","RNTO","ABOR",
"DELE","CWD","LIST","NLST","SITE","STAT","HELP","NOOP","MKD","RMD","PWD","CDUP",
"STOU","SMNT","SYST","SIZE","MDTM","LPSV","LPRT","SPASV","SPORT","UMASK","IDLE",
"CHMOD","LEXERR","STRING","NUMBER","cmd_list","cmd","rcmd","username","password",
"byte_size","host_port","form_code","type_code","struct_code","mode_code","pathname",
"pathstring","octal_number","check_login", NULL
};
#endif

static const short yyr1[] = {     0,
    70,    70,    70,    71,    71,    71,    71,    71,    71,    71,
    71,    71,    71,    71,    71,    71,    71,    71,    71,    71,
    71,    71,    71,    71,    71,    71,    71,    71,    71,    71,
    71,    71,    71,    71,    71,    71,    71,    71,    71,    71,
    71,    71,    71,    71,    71,    71,    71,    71,    71,    71,
    72,    72,    73,    74,    74,    75,    76,    77,    77,    77,
    78,    78,    78,    78,    78,    78,    78,    78,    79,    79,
    79,    80,    80,    80,    81,    82,    83,    84
};

static const short yyr2[] = {     0,
     0,     2,     2,     4,     4,     5,     5,     5,     2,     2,
     2,     4,     4,     4,     4,     8,     5,     5,     5,     3,
     5,     3,     5,     5,     2,     5,     4,     2,     3,     5,
     2,     4,     2,     5,     5,     3,     3,     4,     6,     5,
     7,     9,     4,     6,     5,     2,     5,     5,     2,     2,
     5,     4,     1,     0,     1,     1,    11,     1,     1,     1,
     1,     3,     1,     3,     1,     1,     3,     2,     1,     1,
     1,     1,     1,     1,     1,     1,     1,     0
};

static const short yydefact[] = {     1,
     0,     0,     0,     0,     0,    78,     0,     0,     0,     0,
    78,    78,    78,     0,     0,    78,     0,     0,    78,    78,
    78,    78,     0,    78,     0,     0,    78,    78,    78,    78,
    78,     0,    78,    78,     0,    78,     0,    78,     2,     3,
    50,     0,    54,    49,     0,     9,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,    28,     0,     0,     0,
     0,     0,    25,     0,     0,    31,    33,     0,     0,     0,
     0,     0,    46,     0,     0,    11,     0,    10,     0,    53,
     0,    55,     0,     0,    61,    63,    65,    66,     0,    69,
    71,    70,     0,    73,    74,    72,     0,     0,     0,     0,
     0,    56,     0,     0,    76,     0,    75,     0,     0,    29,
     0,    22,     0,    20,     0,    78,     0,    78,     0,     0,
     0,     0,    36,    37,     0,     0,     0,     0,     0,     4,
     5,     0,     0,     0,     0,     0,    68,    12,    13,    14,
     0,     0,     0,     0,    15,    52,     0,    27,     0,     0,
     0,     0,     0,    38,     0,     0,    43,     0,     0,    32,
     0,     0,     0,     0,     0,     0,     0,     0,     6,    60,
    58,    59,    62,    64,    67,    17,    18,    19,     0,    51,
    26,    30,    23,    21,     0,     0,    40,     0,     0,    24,
    34,    35,    45,    47,    48,     8,     7,     0,     0,    39,
    77,     0,    44,     0,     0,     0,    41,     0,     0,    16,
     0,     0,    42,     0,     0,     0,     0,    57,     0,     0
};

static const short yydefgoto[] = {     1,
    39,    40,    81,    83,   103,   133,   173,    89,    93,    97,
   106,   107,   202,    45
};

static const short yypact[] = {-32768,
    46,     9,    29,    68,    86,-32768,    97,    99,   100,   101,
-32768,-32768,-32768,   102,   103,-32768,   104,   105,-32768,-32768,
-32768,-32768,   107,   108,    20,   109,-32768,-32768,-32768,-32768,
-32768,   110,-32768,-32768,   111,-32768,   112,-32768,-32768,-32768,
-32768,    52,    55,-32768,   114,-32768,     6,    49,     3,   115,
   116,   117,    65,    67,   118,    70,-32768,   120,    23,    47,
    51,   -46,-32768,   125,    73,-32768,-32768,   127,   128,   129,
   130,   132,-32768,   133,   134,-32768,   135,-32768,   136,-32768,
   137,-32768,   138,    75,   140,   141,-32768,   -11,   142,-32768,
-32768,-32768,   143,-32768,-32768,-32768,   144,    70,    70,    70,
    62,-32768,   145,    70,-32768,   146,-32768,    70,    70,-32768,
    70,-32768,    84,-32768,    64,-32768,    66,-32768,    70,   147,
    70,    70,-32768,-32768,    70,    70,    70,    70,    70,-32768,
-32768,   148,   150,    12,    12,    67,-32768,-32768,-32768,-32768,
   151,   152,   153,   158,-32768,-32768,   155,-32768,   156,   157,
   159,   160,    89,-32768,    96,    95,-32768,   162,   163,-32768,
   164,   165,   166,   167,   168,   169,   170,   119,-32768,-32768,
-32768,-32768,-32768,-32768,-32768,-32768,-32768,-32768,   172,-32768,
-32768,-32768,-32768,-32768,   173,   121,-32768,   175,   121,-32768,
-32768,-32768,-32768,-32768,-32768,-32768,-32768,   161,   123,-32768,
-32768,   177,-32768,   179,   126,   180,-32768,    70,   181,-32768,
   183,   131,-32768,   184,   139,   185,   149,-32768,   174,-32768
};

static const short yypgoto[] = {-32768,
-32768,-32768,-32768,-32768,   -83,-32768,    69,-32768,-32768,-32768,
   -98,-32768,     8,    21
};


#define	YYLAST		218


static const short yytable[] = {   141,
   142,   143,   115,   136,   137,   147,    94,    95,    85,   149,
   150,    86,   151,    87,    88,    96,   170,   116,   117,   118,
   159,   171,   161,   162,    41,   172,   163,   164,   165,   166,
   167,    50,    51,    52,    65,    66,    55,   109,   110,    58,
    59,    60,    61,    42,    64,   219,     2,    68,    69,    70,
    71,    72,   175,    74,    75,    90,    77,   102,    79,    91,
    92,   111,   112,     3,     4,   113,   114,     5,     6,     7,
     8,     9,    10,    11,    12,    13,   144,   145,   153,   154,
   156,   157,    43,    14,    15,    16,    17,    18,    19,    20,
    21,    22,    23,    24,    25,    26,    27,    28,    29,    30,
    31,    44,    32,    33,    34,    35,    36,    37,    38,   211,
   186,   187,    46,    47,    48,    49,    53,    54,    56,    80,
    57,    62,    82,    63,    67,    73,    76,    78,    84,    98,
    99,   100,   104,   101,   108,   102,   155,   105,   158,   119,
   120,   121,   122,   132,   123,   124,   125,   126,   127,   128,
   129,   152,   130,   131,   134,   135,   185,   138,   139,   140,
   146,   148,   160,   188,   168,   169,   176,   177,   178,   179,
   180,   181,   182,   220,   183,   184,   189,   205,   190,   191,
   192,   193,   194,   195,   196,   197,   199,   198,   200,   201,
   203,   206,   207,   208,   209,   210,   204,   212,   213,   214,
   215,   217,     0,   174,     0,     0,     0,   216,     0,     0,
     0,     0,     0,     0,     0,     0,     0,   218
};

static const short yycheck[] = {    98,
    99,   100,    49,    15,    88,   104,     4,     5,     3,   108,
   109,     6,   111,     8,     9,    13,     5,    64,    65,    66,
   119,    10,   121,   122,    16,    14,   125,   126,   127,   128,
   129,    11,    12,    13,    15,    16,    16,    15,    16,    19,
    20,    21,    22,    15,    24,     0,     1,    27,    28,    29,
    30,    31,   136,    33,    34,     7,    36,    69,    38,    11,
    12,    15,    16,    18,    19,    15,    16,    22,    23,    24,
    25,    26,    27,    28,    29,    30,    15,    16,    15,    16,
    15,    16,    15,    38,    39,    40,    41,    42,    43,    44,
    45,    46,    47,    48,    49,    50,    51,    52,    53,    54,
    55,    16,    57,    58,    59,    60,    61,    62,    63,   208,
    15,    16,    16,    15,    15,    15,    15,    15,    15,    68,
    16,    15,    68,    16,    16,    16,    16,    16,    15,    15,
    15,    15,    15,    69,    15,    69,   116,    68,   118,    15,
    68,    15,    15,    69,    16,    16,    15,    15,    15,    15,
    15,    68,    16,    16,    15,    15,    68,    16,    16,    16,
    16,    16,    16,    69,    17,    16,    16,    16,    16,    12,
    16,    16,    16,     0,    16,    16,    15,    17,    16,    16,
    16,    16,    16,    16,    16,    16,    15,    69,    16,    69,
    16,    69,    16,    15,    69,    16,   189,    17,    16,    69,
    17,    17,    -1,   135,    -1,    -1,    -1,    69,    -1,    -1,
    -1,    -1,    -1,    -1,    -1,    -1,    -1,    69
};
/* -*-C-*-  Note some compilers choke on comments on `#line' lines.  */
#line 3 "/usr/local/share/bison.simple"

/* Skeleton output parser for bison,
   Copyright (C) 1984, 1989, 1990 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* As a special exception, when this file is copied by Bison into a
   Bison output file, you may use that output file without restriction.
   This special exception was added by the Free Software Foundation
   in version 1.24 of Bison.  */

#ifndef alloca
#ifdef __GNUC__
#define alloca __builtin_alloca
#else /* not GNU C.  */
#if (!defined (__STDC__) && defined (sparc)) || defined (__sparc__) || defined (__sparc) || defined (__sgi)
#include <alloca.h>
#else /* not sparc */
#if defined (MSDOS) && !defined (__TURBOC__)
#include <malloc.h>
#else /* not MSDOS, or __TURBOC__ */
#if defined(_AIX)
#include <malloc.h>
 #pragma alloca
#else /* not MSDOS, __TURBOC__, or _AIX */
#ifdef __hpux
#ifdef __cplusplus
extern "C" {
void *alloca (unsigned int);
};
#else /* not __cplusplus */
void *alloca ();
#endif /* not __cplusplus */
#endif /* __hpux */
#endif /* not _AIX */
#endif /* not MSDOS, or __TURBOC__ */
#endif /* not sparc.  */
#endif /* not GNU C.  */
#endif /* alloca not defined.  */

/* This is the parser code that is written into each bison parser
  when the %semantic_parser declaration is not specified in the grammar.
  It was written by Richard Stallman by simplifying the hairy parser
  used when %semantic_parser is specified.  */

/* Note: there must be only one dollar sign in this file.
   It is replaced by the list of actions, each action
   as one case of the switch.  */

#define yyerrok		(yyerrstatus = 0)
#define yyclearin	(yychar = YYEMPTY)
#define YYEMPTY		-2
#define YYEOF		0
#define YYACCEPT	return(0)
#define YYABORT 	return(1)
#define YYERROR		goto yyerrlab1
/* Like YYERROR except do call yyerror.
   This remains here temporarily to ease the
   transition to the new meaning of YYERROR, for GCC.
   Once GCC version 2 has supplanted version 1, this can go.  */
#define YYFAIL		goto yyerrlab
#define YYRECOVERING()  (!!yyerrstatus)
#define YYBACKUP(token, value) \
do								\
  if (yychar == YYEMPTY && yylen == 1)				\
    { yychar = (token), yylval = (value);			\
      yychar1 = YYTRANSLATE (yychar);				\
      YYPOPSTACK;						\
      goto yybackup;						\
    }								\
  else								\
    { yyerror ("syntax error: cannot back up"); YYERROR; }	\
while (0)

#define YYTERROR	1
#define YYERRCODE	256

#ifndef YYPURE
#define YYLEX		yylex()
#endif

#ifdef YYPURE
#ifdef YYLSP_NEEDED
#ifdef YYLEX_PARAM
#define YYLEX		yylex(&yylval, &yylloc, YYLEX_PARAM)
#else
#define YYLEX		yylex(&yylval, &yylloc)
#endif
#else /* not YYLSP_NEEDED */
#ifdef YYLEX_PARAM
#define YYLEX		yylex(&yylval, YYLEX_PARAM)
#else
#define YYLEX		yylex(&yylval)
#endif
#endif /* not YYLSP_NEEDED */
#endif

/* If nonreentrant, generate the variables here */

#ifndef YYPURE

int	yychar;			/*  the lookahead symbol		*/
YYSTYPE	yylval;			/*  the semantic value of the		*/
				/*  lookahead symbol			*/

#ifdef YYLSP_NEEDED
YYLTYPE yylloc;			/*  location data for the lookahead	*/
				/*  symbol				*/
#endif

int yynerrs;			/*  number of parse errors so far       */
#endif  /* not YYPURE */

#if YYDEBUG != 0
int yydebug;			/*  nonzero means print parse trace	*/
/* Since this is uninitialized, it does not stop multiple parsers
   from coexisting.  */
#endif

/*  YYINITDEPTH indicates the initial size of the parser's stacks	*/

#ifndef	YYINITDEPTH
#define YYINITDEPTH 200
#endif

/*  YYMAXDEPTH is the maximum size the stacks can grow to
    (effective only if the built-in stack extension method is used).  */

#if YYMAXDEPTH == 0
#undef YYMAXDEPTH
#endif

#ifndef YYMAXDEPTH
#define YYMAXDEPTH 10000
#endif

/* Prevent warning if -Wstrict-prototypes.  */
#ifdef __GNUC__
int yyparse (void);
#endif

#if __GNUC__ > 1		/* GNU C and GNU C++ define this.  */
#define __yy_memcpy(TO,FROM,COUNT)	__builtin_memcpy(TO,FROM,COUNT)
#else				/* not GNU C or C++ */
#ifndef __cplusplus

/* This is the most reliable way to avoid incompatibilities
   in available built-in functions on various systems.  */
static void
__yy_memcpy (to, from, count)
     char *to;
     char *from;
     int count;
{
  register char *f = from;
  register char *t = to;
  register int i = count;

  while (i-- > 0)
    *t++ = *f++;
}

#else /* __cplusplus */

/* This is the most reliable way to avoid incompatibilities
   in available built-in functions on various systems.  */
static void
__yy_memcpy (char *to, char *from, int count)
{
  register char *f = from;
  register char *t = to;
  register int i = count;

  while (i-- > 0)
    *t++ = *f++;
}

#endif
#endif

#line 196 "/usr/local/share/bison.simple"

/* The user can define YYPARSE_PARAM as the name of an argument to be passed
   into yyparse.  The argument should have type void *.
   It should actually point to an object.
   Grammar actions can access the variable by casting it
   to the proper pointer type.  */

#ifdef YYPARSE_PARAM
#ifdef __cplusplus
#define YYPARSE_PARAM_ARG void *YYPARSE_PARAM
#define YYPARSE_PARAM_DECL
#else /* not __cplusplus */
#define YYPARSE_PARAM_ARG YYPARSE_PARAM
#define YYPARSE_PARAM_DECL void *YYPARSE_PARAM;
#endif /* not __cplusplus */
#else /* not YYPARSE_PARAM */
#define YYPARSE_PARAM_ARG
#define YYPARSE_PARAM_DECL
#endif /* not YYPARSE_PARAM */

int
yyparse(YYPARSE_PARAM_ARG)
     YYPARSE_PARAM_DECL
{
  register int yystate;
  register int yyn;
  register short *yyssp;
  register YYSTYPE *yyvsp;
  int yyerrstatus;	/*  number of tokens to shift before error messages enabled */
  int yychar1 = 0;		/*  lookahead token as an internal (translated) token number */

  short	yyssa[YYINITDEPTH];	/*  the state stack			*/
  YYSTYPE yyvsa[YYINITDEPTH];	/*  the semantic value stack		*/

  short *yyss = yyssa;		/*  refer to the stacks thru separate pointers */
  YYSTYPE *yyvs = yyvsa;	/*  to allow yyoverflow to reallocate them elsewhere */

#ifdef YYLSP_NEEDED
  YYLTYPE yylsa[YYINITDEPTH];	/*  the location stack			*/
  YYLTYPE *yyls = yylsa;
  YYLTYPE *yylsp;

#define YYPOPSTACK   (yyvsp--, yyssp--, yylsp--)
#else
#define YYPOPSTACK   (yyvsp--, yyssp--)
#endif

  int yystacksize = YYINITDEPTH;

#ifdef YYPURE
  int yychar;
  YYSTYPE yylval;
  int yynerrs;
#ifdef YYLSP_NEEDED
  YYLTYPE yylloc;
#endif
#endif

  YYSTYPE yyval;		/*  the variable used to return		*/
				/*  semantic values from the action	*/
				/*  routines				*/

  int yylen;

#if YYDEBUG != 0
  if (yydebug)
    fprintf(stderr, "Starting parse\n");
#endif

  yystate = 0;
  yyerrstatus = 0;
  yynerrs = 0;
  yychar = YYEMPTY;		/* Cause a token to be read.  */

  /* Initialize stack pointers.
     Waste one element of value and location stack
     so that they stay on the same level as the state stack.
     The wasted elements are never initialized.  */

  yyssp = yyss - 1;
  yyvsp = yyvs;
#ifdef YYLSP_NEEDED
  yylsp = yyls;
#endif

/* Push a new state, which is found in  yystate  .  */
/* In all cases, when you get here, the value and location stacks
   have just been pushed. so pushing a state here evens the stacks.  */
yynewstate:

  *++yyssp = yystate;

  if (yyssp >= yyss + yystacksize - 1)
    {
      /* Give user a chance to reallocate the stack */
      /* Use copies of these so that the &'s don't force the real ones into memory. */
      YYSTYPE *yyvs1 = yyvs;
      short *yyss1 = yyss;
#ifdef YYLSP_NEEDED
      YYLTYPE *yyls1 = yyls;
#endif

      /* Get the current used size of the three stacks, in elements.  */
      int size = yyssp - yyss + 1;

#ifdef yyoverflow
      /* Each stack pointer address is followed by the size of
	 the data in use in that stack, in bytes.  */
#ifdef YYLSP_NEEDED
      /* This used to be a conditional around just the two extra args,
	 but that might be undefined if yyoverflow is a macro.  */
      yyoverflow("parser stack overflow",
		 &yyss1, size * sizeof (*yyssp),
		 &yyvs1, size * sizeof (*yyvsp),
		 &yyls1, size * sizeof (*yylsp),
		 &yystacksize);
#else
      yyoverflow("parser stack overflow",
		 &yyss1, size * sizeof (*yyssp),
		 &yyvs1, size * sizeof (*yyvsp),
		 &yystacksize);
#endif

      yyss = yyss1; yyvs = yyvs1;
#ifdef YYLSP_NEEDED
      yyls = yyls1;
#endif
#else /* no yyoverflow */
      /* Extend the stack our own way.  */
      if (yystacksize >= YYMAXDEPTH)
	{
	  yyerror("parser stack overflow");
	  return 2;
	}
      yystacksize *= 2;
      if (yystacksize > YYMAXDEPTH)
	yystacksize = YYMAXDEPTH;
      yyss = (short *) alloca (yystacksize * sizeof (*yyssp));
      __yy_memcpy ((char *)yyss, (char *)yyss1, size * sizeof (*yyssp));
      yyvs = (YYSTYPE *) alloca (yystacksize * sizeof (*yyvsp));
      __yy_memcpy ((char *)yyvs, (char *)yyvs1, size * sizeof (*yyvsp));
#ifdef YYLSP_NEEDED
      yyls = (YYLTYPE *) alloca (yystacksize * sizeof (*yylsp));
      __yy_memcpy ((char *)yyls, (char *)yyls1, size * sizeof (*yylsp));
#endif
#endif /* no yyoverflow */

      yyssp = yyss + size - 1;
      yyvsp = yyvs + size - 1;
#ifdef YYLSP_NEEDED
      yylsp = yyls + size - 1;
#endif

#if YYDEBUG != 0
      if (yydebug)
	fprintf(stderr, "Stack size increased to %d\n", yystacksize);
#endif

      if (yyssp >= yyss + yystacksize - 1)
	YYABORT;
    }

#if YYDEBUG != 0
  if (yydebug)
    fprintf(stderr, "Entering state %d\n", yystate);
#endif

  goto yybackup;
 yybackup:

/* Do appropriate processing given the current state.  */
/* Read a lookahead token if we need one and don't already have one.  */
/* yyresume: */

  /* First try to decide what to do without reference to lookahead token.  */

  yyn = yypact[yystate];
  if (yyn == YYFLAG)
    goto yydefault;

  /* Not known => get a lookahead token if don't already have one.  */

  /* yychar is either YYEMPTY or YYEOF
     or a valid token in external form.  */

  if (yychar == YYEMPTY)
    {
#if YYDEBUG != 0
      if (yydebug)
	fprintf(stderr, "Reading a token: ");
#endif
      yychar = YYLEX;
    }

  /* Convert token to internal form (in yychar1) for indexing tables with */

  if (yychar <= 0)		/* This means end of input. */
    {
      yychar1 = 0;
      yychar = YYEOF;		/* Don't call YYLEX any more */

#if YYDEBUG != 0
      if (yydebug)
	fprintf(stderr, "Now at end of input.\n");
#endif
    }
  else
    {
      yychar1 = YYTRANSLATE(yychar);

#if YYDEBUG != 0
      if (yydebug)
	{
	  fprintf (stderr, "Next token is %d (%s", yychar, yytname[yychar1]);
	  /* Give the individual parser a way to print the precise meaning
	     of a token, for further debugging info.  */
#ifdef YYPRINT
	  YYPRINT (stderr, yychar, yylval);
#endif
	  fprintf (stderr, ")\n");
	}
#endif
    }

  yyn += yychar1;
  if (yyn < 0 || yyn > YYLAST || yycheck[yyn] != yychar1)
    goto yydefault;

  yyn = yytable[yyn];

  /* yyn is what to do for this token type in this state.
     Negative => reduce, -yyn is rule number.
     Positive => shift, yyn is new state.
       New state is final state => don't bother to shift,
       just return success.
     0, or most negative number => error.  */

  if (yyn < 0)
    {
      if (yyn == YYFLAG)
	goto yyerrlab;
      yyn = -yyn;
      goto yyreduce;
    }
  else if (yyn == 0)
    goto yyerrlab;

  if (yyn == YYFINAL)
    YYACCEPT;

  /* Shift the lookahead token.  */

#if YYDEBUG != 0
  if (yydebug)
    fprintf(stderr, "Shifting token %d (%s), ", yychar, yytname[yychar1]);
#endif

  /* Discard the token being shifted unless it is eof.  */
  if (yychar != YYEOF)
    yychar = YYEMPTY;

  *++yyvsp = yylval;
#ifdef YYLSP_NEEDED
  *++yylsp = yylloc;
#endif

  /* count tokens shifted since error; after three, turn off error status.  */
  if (yyerrstatus) yyerrstatus--;

  yystate = yyn;
  goto yynewstate;

/* Do the default action for the current state.  */
yydefault:

  yyn = yydefact[yystate];
  if (yyn == 0)
    goto yyerrlab;

/* Do a reduction.  yyn is the number of a rule to reduce with.  */
yyreduce:
  yylen = yyr2[yyn];
  if (yylen > 0)
    yyval = yyvsp[1-yylen]; /* implement default value of the action */

#if YYDEBUG != 0
  if (yydebug)
    {
      int i;

      fprintf (stderr, "Reducing via rule %d (line %d), ",
	       yyn, yyrline[yyn]);

      /* Print the symbols being reduced, and their result.  */
      for (i = yyprhs[yyn]; yyrhs[i] > 0; i++)
	fprintf (stderr, "%s ", yytname[yyrhs[i]]);
      fprintf (stderr, " -> %s\n", yytname[yyr1[yyn]]);
    }
#endif


  switch (yyn) {

case 2:
#line 165 "ftpcmd.y"
{
			fromname = (char *) 0;
			restart_point = (off_t) 0;
		;
    break;}
case 4:
#line 174 "ftpcmd.y"
{
			user(yyvsp[-1].s);
			free(yyvsp[-1].s);
		;
    break;}
case 5:
#line 179 "ftpcmd.y"
{
			pass(yyvsp[-1].s);
			free(yyvsp[-1].s);
		;
    break;}
case 6:
#line 184 "ftpcmd.y"
{
			usedefault = 0;
			if (pdata >= 0) {
				(void) close(pdata);
				pdata = -1;
			}
			if (yyvsp[-3].i && yyvsp[-1].i)
			  reply(200, "PORT command successful.");
			else
			  syslog(LOG_WARNING, "refused %s from %s", cbuf, remotehost);
		;
    break;}
case 7:
#line 196 "ftpcmd.y"
{
#ifdef INET6
		usedefault = 0;
		if (pdata >= 0) {
		  (void) close(pdata);
		  pdata = -1;
		}
		if (yyvsp[-3].i) {
		  if (logged_in) {
		    switch(sporttosa(&data_dest, yyvsp[-1].s, &his_addr)) {
		      case 0:
		        reply(200, "SPORT command successful.");
                        break;
                      case -3:
                        reply(501, "Address invalid for your control connection", nrl_afnumtoname(ctrl_addr.sa.sa_family));
		        syslog(LOG_WARNING, "refused bogus %s from %s", cbuf, remotehost);
		        break;
                      default:
		        reply(501, "Address syntax invalid");
                    }
		  }
		} else
		  syslog(LOG_WARNING, "refused bogus %s from %s", cbuf, remotehost);
#endif /* INET6 */
		;
    break;}
case 8:
#line 222 "ftpcmd.y"
{
#ifdef INET6
		usedefault = 0;
		if (pdata >= 0) {
		  (void) close(pdata);
		  pdata = -1;
		}
		if (yyvsp[-3].i) {
		  if (logged_in) {
		    switch(lporttosa(&data_dest, yyvsp[-1].s, &his_addr)) {
		      case 0:
		        reply(200, "LPRT command successful.");
                        break;
                      case -2:
                        reply(501, "Address invalid for a %s control connection", nrl_afnumtoname(ctrl_addr.sa.sa_family));
		        break;
                      case -3:
                        reply(501, "Address invalid for your control connection", nrl_afnumtoname(ctrl_addr.sa.sa_family));
		        syslog(LOG_WARNING, "refused bogus %s from %s", cbuf, remotehost);
		        break;
                      default:
		        reply(501, "Address syntax invalid");
                    }
		  }
		} else
		  syslog(LOG_WARNING, "refused bogus %s from %s", cbuf, remotehost);
#endif /* INET6 */
		;
    break;}
case 9:
#line 251 "ftpcmd.y"
{
#ifdef INET6
			passive(2);
#else /* INET6 */
			passive();
#endif /* INET6 */
		;
    break;}
case 10:
#line 259 "ftpcmd.y"
{
#ifdef INET6
			passive(1);
#endif /* INET6 */
		;
    break;}
case 11:
#line 265 "ftpcmd.y"
{
#ifdef INET6
			passive(3);
#endif /* INET6 */
		;
    break;}
case 12:
#line 271 "ftpcmd.y"
{
			switch (cmd_type) {

			case TYPE_A:
				if (cmd_form == FORM_N) {
					reply(200, "Type set to A.");
					type = cmd_type;
					form = cmd_form;
				} else
					reply(504, "Form must be N.");
				break;

			case TYPE_E:
				reply(504, "Type E not implemented.");
				break;

			case TYPE_I:
				reply(200, "Type set to I.");
				type = cmd_type;
				break;

			case TYPE_L:
#ifndef NBBY
#define NBBY 8
#endif /* NBBY */
#if NBBY == 8
				if (cmd_bytesz == 8) {
					reply(200,
					    "Type set to L (byte size 8).");
					type = cmd_type;
				} else
					reply(504, "Byte size must be 8.");
#else /* NBBY == 8 */
				UNIMPLEMENTED for NBBY != 8
#endif /* NBBY == 8 */
			}
		;
    break;}
case 13:
#line 309 "ftpcmd.y"
{
			switch (yyvsp[-1].i) {

			case STRU_F:
				reply(200, "STRU F ok.");
				break;

			default:
				reply(504, "Unimplemented STRU type.");
			}
		;
    break;}
case 14:
#line 321 "ftpcmd.y"
{
			switch (yyvsp[-1].i) {

			case MODE_S:
				reply(200, "MODE S ok.");
				break;

			default:
				reply(502, "Unimplemented MODE type.");
			}
		;
    break;}
case 15:
#line 333 "ftpcmd.y"
{
			reply(202, "ALLO command ignored.");
		;
    break;}
case 16:
#line 337 "ftpcmd.y"
{
			reply(202, "ALLO command ignored.");
		;
    break;}
case 17:
#line 341 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				retrieve((char *) 0, yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 18:
#line 348 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				store(yyvsp[-1].s, "w", 0);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 19:
#line 355 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				store(yyvsp[-1].s, "a", 0);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 20:
#line 362 "ftpcmd.y"
{
			if (yyvsp[-1].i)
				send_file_list(".");
		;
    break;}
case 21:
#line 367 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				send_file_list(yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 22:
#line 374 "ftpcmd.y"
{
			if (yyvsp[-1].i)
				retrieve("/bin/ls -lgA", "");
		;
    break;}
case 23:
#line 379 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				retrieve("/bin/ls -lgA %s", yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 24:
#line 386 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				statfilecmd(yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 25:
#line 393 "ftpcmd.y"
{
			statcmd();
		;
    break;}
case 26:
#line 397 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				delete(yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 27:
#line 404 "ftpcmd.y"
{
			if (fromname) {
				renamecmd(fromname, yyvsp[-1].s);
				free(fromname);
				fromname = (char *) 0;
			} else {
				reply(503, "Bad sequence of commands.");
			}
			free(yyvsp[-1].s);
		;
    break;}
case 28:
#line 415 "ftpcmd.y"
{
			reply(225, "ABOR command successful.");
		;
    break;}
case 29:
#line 419 "ftpcmd.y"
{
			if (yyvsp[-1].i)
				cwd(pw->pw_dir);
		;
    break;}
case 30:
#line 424 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				cwd(yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 31:
#line 431 "ftpcmd.y"
{
			help(cmdtab, (char *) 0);
		;
    break;}
case 32:
#line 435 "ftpcmd.y"
{
			char *cp = yyvsp[-1].s;

			if (strncasecmp(cp, "SITE", 4) == 0) {
				cp = yyvsp[-1].s + 4;
				if (*cp == ' ')
					cp++;
				if (*cp)
					help(sitetab, cp);
				else
					help(sitetab, (char *) 0);
			} else
				help(cmdtab, yyvsp[-1].s);
		;
    break;}
case 33:
#line 450 "ftpcmd.y"
{
			reply(200, "NOOP command successful.");
		;
    break;}
case 34:
#line 454 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				makedir(yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 35:
#line 461 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				removedir(yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 36:
#line 468 "ftpcmd.y"
{
			if (yyvsp[-1].i)
				pwd();
		;
    break;}
case 37:
#line 473 "ftpcmd.y"
{
			if (yyvsp[-1].i)
				cwd("..");
		;
    break;}
case 38:
#line 478 "ftpcmd.y"
{
			help(sitetab, (char *) 0);
		;
    break;}
case 39:
#line 482 "ftpcmd.y"
{
			help(sitetab, yyvsp[-1].s);
		;
    break;}
case 40:
#line 486 "ftpcmd.y"
{
			int oldmask;

			if (yyvsp[-1].i) {
				oldmask = umask(0);
				(void) umask(oldmask);
				reply(200, "Current UMASK is %03o", oldmask);
			}
		;
    break;}
case 41:
#line 496 "ftpcmd.y"
{
			int oldmask;

			if (yyvsp[-3].i) {
				if ((yyvsp[-1].i == -1) || (yyvsp[-1].i > 0777)) {
					reply(501, "Bad UMASK value");
				} else {
					oldmask = umask(yyvsp[-1].i);
					reply(200,
					    "UMASK set to %03o (was %03o)",
					    yyvsp[-1].i, oldmask);
				}
			}
		;
    break;}
case 42:
#line 511 "ftpcmd.y"
{
			if (yyvsp[-5].i && (yyvsp[-1].s != NULL)) {
				if (yyvsp[-3].i > 0777)
					reply(501,
				"CHMOD: Mode value must be between 0 and 0777");
				else if (chmod(yyvsp[-1].s, yyvsp[-3].i) < 0)
					perror_reply(550, yyvsp[-1].s);
				else
					reply(200, "CHMOD command successful.");
			}
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 43:
#line 525 "ftpcmd.y"
{
			reply(200,
			    "Current IDLE time limit is %d seconds; max %d",
				timeout, maxtimeout);
		;
    break;}
case 44:
#line 531 "ftpcmd.y"
{
			if (yyvsp[-1].i < 30 || yyvsp[-1].i > maxtimeout) {
				reply(501,
			"Maximum IDLE time must be between 30 and %d seconds",
				    maxtimeout);
			} else {
				timeout = yyvsp[-1].i;
				(void) alarm((unsigned) timeout);
				reply(200,
				    "Maximum IDLE time set to %d seconds",
				    timeout);
			}
		;
    break;}
case 45:
#line 545 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				store(yyvsp[-1].s, "w", 1);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 46:
#line 552 "ftpcmd.y"
{
#ifdef unix
#ifdef BSD
			reply(215, "UNIX Type: L%d Version: BSD-%d",
				NBBY, BSD);
#else /* BSD */
			reply(215, "UNIX Type: L%d", NBBY);
#endif /* BSD */
#else /* unix */
			reply(215, "UNKNOWN Type: L%d", NBBY);
#endif /* unix */
		;
    break;}
case 47:
#line 573 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL)
				sizecmd(yyvsp[-1].s);
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 48:
#line 590 "ftpcmd.y"
{
			if (yyvsp[-3].i && yyvsp[-1].s != NULL) {
				struct stat stbuf;
				if (stat(yyvsp[-1].s, &stbuf) < 0)
					reply(550, "%s: %s",
					    yyvsp[-1].s, strerror(errno));
				else if (!S_ISREG(stbuf.st_mode)) {
					reply(550, "%s: not a plain file.", yyvsp[-1].s);
				} else {
					struct tm *t;
					t = gmtime(&stbuf.st_mtime);
					reply(213,
					    "19%02d%02d%02d%02d%02d%02d",
					    t->tm_year, t->tm_mon+1, t->tm_mday,
					    t->tm_hour, t->tm_min, t->tm_sec);
				}
			}
			if (yyvsp[-1].s != NULL)
				free(yyvsp[-1].s);
		;
    break;}
case 49:
#line 611 "ftpcmd.y"
{
			reply(221, "Goodbye.");
			dologout(0);
		;
    break;}
case 50:
#line 616 "ftpcmd.y"
{
			yyerrok;
		;
    break;}
case 51:
#line 622 "ftpcmd.y"
{
			char *renamefrom();

			restart_point = (off_t) 0;
			if (yyvsp[-3].i && yyvsp[-1].s) {
				fromname = renamefrom(yyvsp[-1].s);
				if (fromname == (char *) 0 && yyvsp[-1].s) {
					free(yyvsp[-1].s);
				}
			}
		;
    break;}
case 52:
#line 634 "ftpcmd.y"
{
			fromname = (char *) 0;
			restart_point = yyvsp[-1].i;	/* XXX $3 is only "int" */
			reply(350, "Restarting at %qd. %s", restart_point,
			    "Send STORE or RETRIEVE to initiate transfer.");
		;
    break;}
case 54:
#line 648 "ftpcmd.y"
{
			yyval.s = (char *)calloc(1, sizeof(char));
		;
    break;}
case 57:
#line 661 "ftpcmd.y"
{
#ifdef INET6
		  u_int8_t *p;

		  yyval.i = 0;
		  if (logged_in) {
		    memcpy(&data_dest, &his_addr, NRL_SA_LEN((struct sockaddr *)&ctrl_addr));

		    switch(his_addr.sa.sa_family) {
		      case AF_INET:
			p = (u_int8_t *)&data_dest.sin.sin_addr;
			p[0] = yyvsp[-10].i;
			p[1] = yyvsp[-8].i;
			p[2] = yyvsp[-6].i;
			p[3] = yyvsp[-4].i;

			p = (u_int8_t *)&data_dest.sin.sin_port;
			p[0] = yyvsp[-2].i;
			p[1] = yyvsp[0].i;

			if (!memcmp(&data_dest.sin.sin_addr, &his_addr.sin.sin_addr, sizeof(struct in_addr)) && (data_dest.sin.sin_port > 1023))
			  yyval.i = 1;
			break;
		      case AF_INET6:
			if (inet6_isipv4addr(&his_addr.sin6.sin6_addr)) {
			  p = (u_int8_t *)&data_dest.sin6.sin6_addr;
			  p[12] = yyvsp[-10].i;
			  p[13] = yyvsp[-8].i;
			  p[14] = yyvsp[-6].i;
			  p[15] = yyvsp[-4].i;

			  p = (u_int8_t *)&data_dest.sin6.sin6_port;
			  p[0] = yyvsp[-2].i;
			  p[1] = yyvsp[0].i;

			  if (!memcmp(&data_dest.sin6.sin6_addr, &his_addr.sin6.sin6_addr, sizeof(struct in6_addr)) && (data_dest.sin6.sin6_port > 1023))
			    yyval.i = 1;
			  break;
			}
		      default:
			reply(501, "Address can't be used with a %s control connection", nrl_afnumtoname(ctrl_addr.sa.sa_family));
		    }
		  }
#else /* INET6 */
			char *a, *p;

			a = (char *)&data_dest.sin_addr;
			a[0] = yyvsp[-10].i; a[1] = yyvsp[-8].i; a[2] = yyvsp[-6].i; a[3] = yyvsp[-4].i;
			p = (char *)&data_dest.sin_port;
			p[0] = yyvsp[-2].i; p[1] = yyvsp[0].i;
			data_dest.sin_family = AF_INET;
#endif /* INET6 */
		;
    break;}
case 58:
#line 718 "ftpcmd.y"
{
			yyval.i = FORM_N;
		;
    break;}
case 59:
#line 722 "ftpcmd.y"
{
			yyval.i = FORM_T;
		;
    break;}
case 60:
#line 726 "ftpcmd.y"
{
			yyval.i = FORM_C;
		;
    break;}
case 61:
#line 733 "ftpcmd.y"
{
			cmd_type = TYPE_A;
			cmd_form = FORM_N;
		;
    break;}
case 62:
#line 738 "ftpcmd.y"
{
			cmd_type = TYPE_A;
			cmd_form = yyvsp[0].i;
		;
    break;}
case 63:
#line 743 "ftpcmd.y"
{
			cmd_type = TYPE_E;
			cmd_form = FORM_N;
		;
    break;}
case 64:
#line 748 "ftpcmd.y"
{
			cmd_type = TYPE_E;
			cmd_form = yyvsp[0].i;
		;
    break;}
case 65:
#line 753 "ftpcmd.y"
{
			cmd_type = TYPE_I;
		;
    break;}
case 66:
#line 757 "ftpcmd.y"
{
			cmd_type = TYPE_L;
			cmd_bytesz = NBBY;
		;
    break;}
case 67:
#line 762 "ftpcmd.y"
{
			cmd_type = TYPE_L;
			cmd_bytesz = yyvsp[0].i;
		;
    break;}
case 68:
#line 768 "ftpcmd.y"
{
			cmd_type = TYPE_L;
			cmd_bytesz = yyvsp[0].i;
		;
    break;}
case 69:
#line 776 "ftpcmd.y"
{
			yyval.i = STRU_F;
		;
    break;}
case 70:
#line 780 "ftpcmd.y"
{
			yyval.i = STRU_R;
		;
    break;}
case 71:
#line 784 "ftpcmd.y"
{
			yyval.i = STRU_P;
		;
    break;}
case 72:
#line 791 "ftpcmd.y"
{
			yyval.i = MODE_S;
		;
    break;}
case 73:
#line 795 "ftpcmd.y"
{
			yyval.i = MODE_B;
		;
    break;}
case 74:
#line 799 "ftpcmd.y"
{
			yyval.i = MODE_C;
		;
    break;}
case 75:
#line 806 "ftpcmd.y"
{
			/*
			 * Problem: this production is used for all pathname
			 * processing, but only gives a 550 error reply.
			 * This is a valid reply in some cases but not in others.
			 */
			if (logged_in && yyvsp[0].s && *yyvsp[0].s == '~') {
				glob_t gl;
#ifdef GLOB_BRACE
				int flags =
				 GLOB_BRACE|GLOB_NOCHECK|GLOB_QUOTE|GLOB_TILDE;
#else /* GLOB_BRACE */
				int flags = GLOB_NOCHECK;
#endif /* GLOB_BRACE */

				memset(&gl, 0, sizeof(gl));
				if (glob(yyvsp[0].s, flags, NULL, &gl) ||
				    gl.gl_pathc == 0) {
					reply(550, "not found");
					yyval.s = NULL;
				} else {
					yyval.s = strdup(gl.gl_pathv[0]);
				}
				globfree(&gl);
				free(yyvsp[0].s);
			} else
				yyval.s = yyvsp[0].s;
		;
    break;}
case 77:
#line 842 "ftpcmd.y"
{
			int ret, dec, multby, digit;

			/*
			 * Convert a number that was read as decimal number
			 * to what it would be if it had been read as octal.
			 */
			dec = yyvsp[0].i;
			multby = 1;
			ret = 0;
			while (dec) {
				digit = dec%10;
				if (digit > 7) {
					ret = -1;
					break;
				}
				ret += digit * multby;
				multby *= 8;
				dec /= 10;
			}
			yyval.i = ret;
		;
    break;}
case 78:
#line 869 "ftpcmd.y"
{
			if (logged_in)
				yyval.i = 1;
			else {
				reply(530, "Please login with USER and PASS.");
				yyval.i = 0;
			}
		;
    break;}
}
   /* the action file gets copied in in place of this dollarsign */
#line 498 "/usr/local/share/bison.simple"

  yyvsp -= yylen;
  yyssp -= yylen;
#ifdef YYLSP_NEEDED
  yylsp -= yylen;
#endif

#if YYDEBUG != 0
  if (yydebug)
    {
      short *ssp1 = yyss - 1;
      fprintf (stderr, "state stack now");
      while (ssp1 != yyssp)
	fprintf (stderr, " %d", *++ssp1);
      fprintf (stderr, "\n");
    }
#endif

  *++yyvsp = yyval;

#ifdef YYLSP_NEEDED
  yylsp++;
  if (yylen == 0)
    {
      yylsp->first_line = yylloc.first_line;
      yylsp->first_column = yylloc.first_column;
      yylsp->last_line = (yylsp-1)->last_line;
      yylsp->last_column = (yylsp-1)->last_column;
      yylsp->text = 0;
    }
  else
    {
      yylsp->last_line = (yylsp+yylen-1)->last_line;
      yylsp->last_column = (yylsp+yylen-1)->last_column;
    }
#endif

  /* Now "shift" the result of the reduction.
     Determine what state that goes to,
     based on the state we popped back to
     and the rule number reduced by.  */

  yyn = yyr1[yyn];

  yystate = yypgoto[yyn - YYNTBASE] + *yyssp;
  if (yystate >= 0 && yystate <= YYLAST && yycheck[yystate] == *yyssp)
    yystate = yytable[yystate];
  else
    yystate = yydefgoto[yyn - YYNTBASE];

  goto yynewstate;

yyerrlab:   /* here on detecting error */

  if (! yyerrstatus)
    /* If not already recovering from an error, report this error.  */
    {
      ++yynerrs;

#ifdef YYERROR_VERBOSE
      yyn = yypact[yystate];

      if (yyn > YYFLAG && yyn < YYLAST)
	{
	  int size = 0;
	  char *msg;
	  int x, count;

	  count = 0;
	  /* Start X at -yyn if nec to avoid negative indexes in yycheck.  */
	  for (x = (yyn < 0 ? -yyn : 0);
	       x < (sizeof(yytname) / sizeof(char *)); x++)
	    if (yycheck[x + yyn] == x)
	      size += strlen(yytname[x]) + 15, count++;
	  msg = (char *) malloc(size + 15);
	  if (msg != 0)
	    {
	      strcpy(msg, "parse error");

	      if (count < 5)
		{
		  count = 0;
		  for (x = (yyn < 0 ? -yyn : 0);
		       x < (sizeof(yytname) / sizeof(char *)); x++)
		    if (yycheck[x + yyn] == x)
		      {
			strcat(msg, count == 0 ? ", expecting `" : " or `");
			strcat(msg, yytname[x]);
			strcat(msg, "'");
			count++;
		      }
		}
	      yyerror(msg);
	      free(msg);
	    }
	  else
	    yyerror ("parse error; also virtual memory exceeded");
	}
      else
#endif /* YYERROR_VERBOSE */
	yyerror("parse error");
    }

  goto yyerrlab1;
yyerrlab1:   /* here on error raised explicitly by an action */

  if (yyerrstatus == 3)
    {
      /* if just tried and failed to reuse lookahead token after an error, discard it.  */

      /* return failure if at end of input */
      if (yychar == YYEOF)
	YYABORT;

#if YYDEBUG != 0
      if (yydebug)
	fprintf(stderr, "Discarding token %d (%s).\n", yychar, yytname[yychar1]);
#endif

      yychar = YYEMPTY;
    }

  /* Else will try to reuse lookahead token
     after shifting the error token.  */

  yyerrstatus = 3;		/* Each real token shifted decrements this */

  goto yyerrhandle;

yyerrdefault:  /* current state does not do anything special for the error token. */

#if 0
  /* This is wrong; only states that explicitly want error tokens
     should shift them.  */
  yyn = yydefact[yystate];  /* If its default is to accept any token, ok.  Otherwise pop it.*/
  if (yyn) goto yydefault;
#endif

yyerrpop:   /* pop the current state because it cannot handle the error token */

  if (yyssp == yyss) YYABORT;
  yyvsp--;
  yystate = *--yyssp;
#ifdef YYLSP_NEEDED
  yylsp--;
#endif

#if YYDEBUG != 0
  if (yydebug)
    {
      short *ssp1 = yyss - 1;
      fprintf (stderr, "Error: state stack now");
      while (ssp1 != yyssp)
	fprintf (stderr, " %d", *++ssp1);
      fprintf (stderr, "\n");
    }
#endif

yyerrhandle:

  yyn = yypact[yystate];
  if (yyn == YYFLAG)
    goto yyerrdefault;

  yyn += YYTERROR;
  if (yyn < 0 || yyn > YYLAST || yycheck[yyn] != YYTERROR)
    goto yyerrdefault;

  yyn = yytable[yyn];
  if (yyn < 0)
    {
      if (yyn == YYFLAG)
	goto yyerrpop;
      yyn = -yyn;
      goto yyreduce;
    }
  else if (yyn == 0)
    goto yyerrpop;

  if (yyn == YYFINAL)
    YYACCEPT;

#if YYDEBUG != 0
  if (yydebug)
    fprintf(stderr, "Shifting error token, ");
#endif

  *++yyvsp = yylval;
#ifdef YYLSP_NEEDED
  *++yylsp = yylloc;
#endif

  yystate = yyn;
  goto yynewstate;
}
#line 879 "ftpcmd.y"


extern jmp_buf errcatch;

#define	CMD	0	/* beginning of command */
#define	ARGS	1	/* expect miscellaneous arguments */
#define	STR1	2	/* expect SP followed by STRING */
#define	STR2	3	/* expect STRING */
#define	OSTR	4	/* optional SP then STRING */
#define	ZSTR1	5	/* SP then optional STRING */
#define	ZSTR2	6	/* optional STRING after SP */
#define	SITECMD	7	/* SITE command */
#define	NSTR	8	/* Number followed by a string */

struct tab cmdtab[] = {		/* In order defined in RFC 765 */
	{ "USER", USER, STR1, 1,	"<sp> username" },
	{ "PASS", PASS, ZSTR1, 1,	"<sp> password" },
	{ "ACCT", ACCT, STR1, 0,	"(specify account)" },
	{ "SMNT", SMNT, ARGS, 0,	"(structure mount)" },
	{ "REIN", REIN, ARGS, 0,	"(reinitialize server state)" },
	{ "QUIT", QUIT, ARGS, 1,	"(terminate service)", },
	{ "PORT", PORT, ARGS, 1,	"<sp> b0, b1, b2, b3, b4" },
	{ "PASV", PASV, ARGS, 1,	"(set server in passive mode)" },
#ifdef INET6
	{ "SPORT", SPORT, STR1, 1,	"<sp> pl, p1, ..." },
	{ "SPASV", SPASV, ARGS, 1,	"(set server in passive mode)" },
	{ "LPRT", LPRT, STR1, 1,	"<sp> af, al, a1, ..., pl, p1, ..." },
	{ "LPSV", LPSV, ARGS, 1,	"(set server in passive mode)" },
#endif /* INET6 */
	{ "TYPE", TYPE, ARGS, 1,	"<sp> [ A | E | I | L ]" },
	{ "STRU", STRU, ARGS, 1,	"(specify file structure)" },
	{ "MODE", MODE, ARGS, 1,	"(specify transfer mode)" },
	{ "RETR", RETR, STR1, 1,	"<sp> file-name" },
	{ "STOR", STOR, STR1, 1,	"<sp> file-name" },
	{ "APPE", APPE, STR1, 1,	"<sp> file-name" },
	{ "MLFL", MLFL, OSTR, 0,	"(mail file)" },
	{ "MAIL", MAIL, OSTR, 0,	"(mail to user)" },
	{ "MSND", MSND, OSTR, 0,	"(mail send to terminal)" },
	{ "MSOM", MSOM, OSTR, 0,	"(mail send to terminal or mailbox)" },
	{ "MSAM", MSAM, OSTR, 0,	"(mail send to terminal and mailbox)" },
	{ "MRSQ", MRSQ, OSTR, 0,	"(mail recipient scheme question)" },
	{ "MRCP", MRCP, STR1, 0,	"(mail recipient)" },
	{ "ALLO", ALLO, ARGS, 1,	"allocate storage (vacuously)" },
	{ "REST", REST, ARGS, 1,	"<sp> offset (restart command)" },
	{ "RNFR", RNFR, STR1, 1,	"<sp> file-name" },
	{ "RNTO", RNTO, STR1, 1,	"<sp> file-name" },
	{ "ABOR", ABOR, ARGS, 1,	"(abort operation)" },
	{ "DELE", DELE, STR1, 1,	"<sp> file-name" },
	{ "CWD",  CWD,  OSTR, 1,	"[ <sp> directory-name ]" },
	{ "XCWD", CWD,	OSTR, 1,	"[ <sp> directory-name ]" },
	{ "LIST", LIST, OSTR, 1,	"[ <sp> path-name ]" },
	{ "NLST", NLST, OSTR, 1,	"[ <sp> path-name ]" },
	{ "SITE", SITE, SITECMD, 1,	"site-cmd [ <sp> arguments ]" },
	{ "SYST", SYST, ARGS, 1,	"(get type of operating system)" },
	{ "STAT", STAT, OSTR, 1,	"[ <sp> path-name ]" },
	{ "HELP", HELP, OSTR, 1,	"[ <sp> <string> ]" },
	{ "NOOP", NOOP, ARGS, 1,	"" },
	{ "MKD",  MKD,  STR1, 1,	"<sp> path-name" },
	{ "XMKD", MKD,  STR1, 1,	"<sp> path-name" },
	{ "RMD",  RMD,  STR1, 1,	"<sp> path-name" },
	{ "XRMD", RMD,  STR1, 1,	"<sp> path-name" },
	{ "PWD",  PWD,  ARGS, 1,	"(return current directory)" },
	{ "XPWD", PWD,  ARGS, 1,	"(return current directory)" },
	{ "CDUP", CDUP, ARGS, 1,	"(change to parent directory)" },
	{ "XCUP", CDUP, ARGS, 1,	"(change to parent directory)" },
	{ "STOU", STOU, STR1, 1,	"<sp> file-name" },
	{ "SIZE", SIZE, OSTR, 1,	"<sp> path-name" },
	{ "MDTM", MDTM, OSTR, 1,	"<sp> path-name" },
	{ NULL,   0,    0,    0,	0 }
};

struct tab sitetab[] = {
	{ "UMASK", UMASK, ARGS, 1,	"[ <sp> umask ]" },
	{ "IDLE", IDLE, ARGS, 1,	"[ <sp> maximum-idle-time ]" },
	{ "CHMOD", CHMOD, NSTR, 1,	"<sp> mode <sp> file-name" },
	{ "HELP", HELP, OSTR, 1,	"[ <sp> <string> ]" },
	{ NULL,   0,    0,    0,	0 }
};

static char	*copy __P((char *));
static struct tab *
		 lookup __P((struct tab *, char *));
static void	 toolong __P((int));
static int	 yylex __P((void));

static struct tab *
lookup(p, cmd)
	struct tab *p;
	char *cmd;
{

	for (; p->name != NULL; p++)
		if (strcmp(cmd, p->name) == 0)
			return (p);
	return (0);
}

#include <arpa/telnet.h>

/*
 * getline - a hacked up version of fgets to ignore TELNET escape codes.
 */
char *
getline(s, n, iop)
	char *s;
	int n;
	FILE *iop;
{
	int c;
	register char *cs;

	cs = s;
/* tmpline may contain saved command from urgent mode interruption */
	for (c = 0; tmpline[c] != '\0' && --n > 0; ++c) {
		*cs++ = tmpline[c];
		if (tmpline[c] == '\n') {
			*cs++ = '\0';
			if (debug)
				syslog(LOG_DEBUG, "command: %s", s);
			tmpline[0] = '\0';
			return(s);
		}
		if (c == 0)
			tmpline[0] = '\0';
	}
	while ((c = getc(iop)) != EOF) {
		c &= 0377;
		if (c == IAC) {
		    if ((c = getc(iop)) != EOF) {
			c &= 0377;
			switch (c) {
			case WILL:
			case WONT:
				c = getc(iop);
				printf("%c%c%c", IAC, DONT, 0377&c);
				(void) fflush(stdout);
				continue;
			case DO:
			case DONT:
				c = getc(iop);
				printf("%c%c%c", IAC, WONT, 0377&c);
				(void) fflush(stdout);
				continue;
			case IAC:
				break;
			default:
				continue;	/* ignore command */
			}
		    }
		}
		*cs++ = c;
		if (--n <= 0 || c == '\n')
			break;
	}
	if (c == EOF && cs == s)
		return (NULL);
	*cs++ = '\0';
	if (debug) {
		if (!guest && strncasecmp("pass ", s, 5) == 0) {
			/* Don't syslog passwords */
			syslog(LOG_DEBUG, "command: %.5s ???", s);
		} else {
			register char *cp;
			register int len;

			/* Don't syslog trailing CR-LF */
			len = strlen(s);
			cp = s + len - 1;
			while (cp >= s && (*cp == '\n' || *cp == '\r')) {
				--cp;
				--len;
			}
			syslog(LOG_DEBUG, "command: %.*s", len, s);
		}
	}
	return (s);
}

static void
toolong(signo)
	int signo;
{

	reply(421,
	    "Timeout (%d seconds): closing control connection.", timeout);
	if (logging)
		syslog(LOG_INFO, "User %s timed out after %d seconds",
		    (pw ? pw -> pw_name : "unknown"), timeout);
	dologout(1);
}

static int
yylex()
{
	static int cpos, state;
	char *cp, *cp2;
	struct tab *p;
	int n;
	char c;

	for (;;) {
		switch (state) {

		case CMD:
			(void) signal(SIGALRM, toolong);
			(void) alarm((unsigned) timeout);
			if (getline(cbuf, sizeof(cbuf)-1, stdin) == NULL) {
				reply(221, "You could at least say goodbye.");
				dologout(0);
			}
			(void) alarm(0);
#ifdef SETPROCTITLE
			if (strncasecmp(cbuf, "PASS", 4) != NULL)
				setproctitle("%s: %s", proctitle, cbuf);
#endif /* SETPROCTITLE */
			if ((cp = strchr(cbuf, '\r'))) {
				*cp++ = '\n';
				*cp = '\0';
			}
			if ((cp = strpbrk(cbuf, " \n")))
				cpos = cp - cbuf;
			if (cpos == 0)
				cpos = 4;
			c = cbuf[cpos];
			cbuf[cpos] = '\0';
			upper(cbuf);
			p = lookup(cmdtab, cbuf);
			cbuf[cpos] = c;
			if (p != 0) {
				if (p->implemented == 0) {
					nack(p->name);
					longjmp(errcatch,0);
					/* NOTREACHED */
				}
				state = p->state;
				yylval.s = p->name;
				return (p->token);
			}
			break;

		case SITECMD:
			if (cbuf[cpos] == ' ') {
				cpos++;
				return (SP);
			}
			cp = &cbuf[cpos];
			if ((cp2 = strpbrk(cp, " \n")))
				cpos = cp2 - cbuf;
			c = cbuf[cpos];
			cbuf[cpos] = '\0';
			upper(cp);
			p = lookup(sitetab, cp);
			cbuf[cpos] = c;
			if (p != 0) {
				if (p->implemented == 0) {
					state = CMD;
					nack(p->name);
					longjmp(errcatch,0);
					/* NOTREACHED */
				}
				state = p->state;
				yylval.s = p->name;
				return (p->token);
			}
			state = CMD;
			break;

		case OSTR:
			if (cbuf[cpos] == '\n') {
				state = CMD;
				return (CRLF);
			}
			/* FALLTHROUGH */

		case STR1:
		case ZSTR1:
		dostr1:
			if (cbuf[cpos] == ' ') {
				cpos++;
				state = state == OSTR ? STR2 : ++state;
				return (SP);
			}
			break;

		case ZSTR2:
			if (cbuf[cpos] == '\n') {
				state = CMD;
				return (CRLF);
			}
			/* FALLTHROUGH */

		case STR2:
			cp = &cbuf[cpos];
			n = strlen(cp);
			cpos += n - 1;
			/*
			 * Make sure the string is nonempty and \n terminated.
			 */
			if (n > 1 && cbuf[cpos] == '\n') {
				cbuf[cpos] = '\0';
				yylval.s = copy(cp);
				cbuf[cpos] = '\n';
				state = ARGS;
				return (STRING);
			}
			break;

		case NSTR:
			if (cbuf[cpos] == ' ') {
				cpos++;
				return (SP);
			}
			if (isdigit(cbuf[cpos])) {
				cp = &cbuf[cpos];
				while (isdigit(cbuf[++cpos]))
					;
				c = cbuf[cpos];
				cbuf[cpos] = '\0';
				yylval.i = atoi(cp);
				cbuf[cpos] = c;
				state = STR1;
				return (NUMBER);
			}
			state = STR1;
			goto dostr1;

		case ARGS:
			if (isdigit(cbuf[cpos])) {
				cp = &cbuf[cpos];
				while (isdigit(cbuf[++cpos]))
					;
				c = cbuf[cpos];
				cbuf[cpos] = '\0';
				yylval.i = atoi(cp);
				cbuf[cpos] = c;
				return (NUMBER);
			}
			switch (cbuf[cpos++]) {

			case '\n':
				state = CMD;
				return (CRLF);

			case ' ':
				return (SP);

			case ',':
				return (COMMA);

			case 'A':
			case 'a':
				return (A);

			case 'B':
			case 'b':
				return (B);

			case 'C':
			case 'c':
				return (C);

			case 'E':
			case 'e':
				return (E);

			case 'F':
			case 'f':
				return (F);

			case 'I':
			case 'i':
				return (I);

			case 'L':
			case 'l':
				return (L);

			case 'N':
			case 'n':
				return (N);

			case 'P':
			case 'p':
				return (P);

			case 'R':
			case 'r':
				return (R);

			case 'S':
			case 's':
				return (S);

			case 'T':
			case 't':
				return (T);

			}
			break;

		default:
			fatal("Unknown state in scanner.");
		}
		yyerror((char *) 0);
		state = CMD;
		longjmp(errcatch,0);
	}
}

void
upper(s)
	char *s;
{
	while (*s != '\0') {
		if (islower(*s))
			*s = toupper(*s);
		s++;
	}
}

static char *
copy(s)
	char *s;
{
	char *p;

	p = malloc((unsigned) strlen(s) + 1);
	if (p == NULL)
		fatal("Ran out of memory.");
	(void) strcpy(p, s);
	return (p);
}

static void
help(ctab, s)
	struct tab *ctab;
	char *s;
{
	struct tab *c;
	int width, NCMDS;
	char *type;

	if (ctab == sitetab)
		type = "SITE ";
	else
		type = "";
	width = 0, NCMDS = 0;
	for (c = ctab; c->name != NULL; c++) {
		int len = strlen(c->name);

		if (len > width)
			width = len;
		NCMDS++;
	}
	width = (width + 8) &~ 7;
	if (s == 0) {
		int i, j, w;
		int columns, lines;

		lreply(214, "The following %scommands are recognized %s.",
		    type, "(* =>'s unimplemented)");
		columns = 76 / width;
		if (columns == 0)
			columns = 1;
		lines = (NCMDS + columns - 1) / columns;
		for (i = 0; i < lines; i++) {
			printf("   ");
			for (j = 0; j < columns; j++) {
				c = ctab + j * lines + i;
				printf("%s%c", c->name,
					c->implemented ? ' ' : '*');
				if (c + lines >= &ctab[NCMDS])
					break;
				w = strlen(c->name) + 1;
				while (w < width) {
					putchar(' ');
					w++;
				}
			}
			printf("\r\n");
		}
		(void) fflush(stdout);
		reply(214, "Direct comments to ftp-bugs@%s.", hostname);
		return;
	}
	upper(s);
	c = lookup(ctab, s);
	if (c == (struct tab *)0) {
		reply(502, "Unknown command %s.", s);
		return;
	}
	if (c->implemented)
		reply(214, "Syntax: %s%s %s", type, c->name, c->help);
	else
		reply(214, "%s%-*s\t%s; unimplemented.", type, width,
		    c->name, c->help);
}

static void
sizecmd(filename)
	char *filename;
{
	switch (type) {
	case TYPE_L:
	case TYPE_I: {
		struct stat stbuf;
		if (stat(filename, &stbuf) < 0 || !S_ISREG(stbuf.st_mode))
			reply(550, "%s: not a plain file.", filename);
		else
			reply(213, "%qu", stbuf.st_size);
		break; }
	case TYPE_A: {
		FILE *fin;
		int c;
		off_t count;
		struct stat stbuf;
		fin = fopen(filename, "r");
		if (fin == NULL) {
			perror_reply(550, filename);
			return;
		}
		if (fstat(fileno(fin), &stbuf) < 0 || !S_ISREG(stbuf.st_mode)) {
			reply(550, "%s: not a plain file.", filename);
			(void) fclose(fin);
			return;
		}

		count = 0;
		while((c=getc(fin)) != EOF) {
			if (c == '\n')	/* will get expanded to \r\n */
				count++;
			count++;
		}
		(void) fclose(fin);

		reply(213, "%qd", count);
		break; }
	default:
		reply(504, "SIZE not implemented for Type %c.", "?AEIL"[type]);
	}
}
